/*

  Compiler settings for MSVC 4.2b:

  (suggested)

  preprocessory defs:  NDEBUG ( not  _DEBUG )
  calling convention: fastcall
  processor: pentium
 
   optimizations:
      Assume No Aliasing
      Global Optimizations 
      Generate Intrinsic Functions
      Favor Fast Code 
      Frame-Pointer Omission 
      Full Optimization

  In-line Function Expansion
    Only __inline

  not using precompiled headers
  not incrementally linked or compiled

*/

#include "first.h"
#include "raster_util.h"
#include "clut.h"
#include "raster.h"
#include <assert.h>
#include "rast_i.h"



using namespace raster;







//
//  2d clipping is performed to 0,0, x_res,y_res rect 
//

void __fastcall trace_edge_clipped(float fx0, float fy0, float fx1,float fy1) 
{  

    int y0 = d_to_int(fy0+0.5f); // round up
    int y1 = d_to_int(fy1+0.5f);
  
    // does line seg exist?
    int dy = y1 - y0;
    if(dy == 0 )
      return;

    float rdx = (fx1 - fx0)/(fy1 - fy0);
	
    // are we a right or a left edge?
    int right;

    if( dy < 0 )
     {
     // we are a right edge
      right = 1;
      dy = -dy;
      swap_obs(fx0,fx1);
      swap_obs(fy0,fy1);
      swap_obs(y0,y1);
     }
    else
     right = 0;

    // trivial y clipping
    if( y0 >= raster::y_res )
    {
       return;
    }
    if( y1 <= 0 )
    {
     return;
    }
    
    if( y0 < 0 ) 
       y0 = 0;
    if( y1 > raster::y_res ) 
       y1 = raster::y_res;

    assert( y1 > y0 );


    // test poly bounds
    if(y0 < y_min) y_min = y0;
    if(y1 > y_max) y_max = y1;


    int dx = d_to_fx16( rdx );

    float pre_step = (float)y0 - fy0;

    int  x    = d_to_fx16( fx0  +  pre_step*rdx ) + (FX1.i()-1);
  

    byte clip_flags = byte( (x   < (FX1.i()/2)    )*1 |
                            (x   > clip_max_x_fx16)*2 |
                            (fx1 < -0.5f          )*4 |
                            (fx1 > clip_max_x_flt )*8 );
                      
    // most common case, all unclipped
    //
    if( clip_flags == 0 )
    {
       for(; (y0<y1) ;  ++y0,x+= dx)
       {
         span_extents[y0][right] = FX1.d(x);
       }
       return;
    }


    // first point off to left.
    if( clip_flags & 1  )
    {
     // step until we are on screen
     for(; ((y0<y1)&&(x < (FX1.i()) ) );  ++y0,x+= dx)
     {
       span_extents[y0][right] = 0;
     }
     if( y0 == y1 )
       return;
    }
    else if( clip_flags & 2 )
    {
     for(; ((y0<y1)&&(x >= clip_max_x_fx16 ) );  ++y0,x+= dx)
     {
       span_extents[y0][right] = raster::x_res;
     }
     if( y0 == y1 )
       return;
    }

    //
    // finish off
    //

    // both inside.
    if( (clip_flags & (4|8) ) == 0  )
    {
       for(; (y0<y1) ;  ++y0,x+= dx)
       {
         span_extents[y0][right] = FX1.d(x);
       }
    }
    else if( clip_flags & 8  )
    {
       for(; ((y0<y1) && (x < clip_max_x_fx16) ) ;  ++y0,x+= dx)
       {
          span_extents[y0][right] = FX1.d(x);
       }

       for(; (y0<y1) ;  ++y0)
       {
          span_extents[y0][right] = raster::x_res;
       }
    }
    else
    {
       for(; ((y0<y1) && (x > 0 ) ) ;  ++y0,x+= dx)
       {
          span_extents[y0][right] = FX1.d(x);
       }

       for(; (y0<y1) ; ++y0)
       {
          span_extents[y0][right] = 0;
       }
    }

} 


//
// don't use this unless you are sure that all verts are in the clipping rect
//
//

void __fastcall trace_edge_unclipped(float fx0, float fy0, float fx1,float fy1) 
{  

    int y0 = d_to_int(fy0+0.5f); // round up
    int y1 = d_to_int(fy1+0.5f);
  
    // does line seg exist?
    int dy = y1 - y0;
    if(dy == 0 )
      return;

    float rdx = (fx1 - fx0)/(fy1 - fy0);
	
    // are we a right or a left edge?
    int right;

    if( dy < 0 )
     {
     // we are a right edge
      right = 1;
      dy = -dy;
      swap_obs(fx0,fx1);
      swap_obs(fy0,fy1);
      swap_obs(y0,y1);
     }
    else
     right = 0;

    // trivial y clipping
    if( y0 >= raster::y_res )
    {
       return;
    }
    if( y1 <= 0 )
    {
     return;
    }
    
    assert( y0 >= 0 );
    assert( y1 <= raster::y_res );
    assert( y1 > y0 );


    // test poly bounds
    if(y0 < y_min) y_min = y0;
    if(y1 > y_max) y_max = y1;


    int dx = d_to_fx16( rdx );

    float pre_step = (float)y0 - fy0;
      
    int  x    = d_to_fx16( fx0 + pre_step*rdx )  + (FX1.i()-1);
  

    if( right )
    {
#define IDX 1

         for(; (y0<y1) ;  ++y0,x+= dx)
         {
          span_extents[y0][IDX] = FX1.d(x);
          assert( span_extents[y0][IDX]  >= 0 );
          assert( span_extents[y0][IDX]  <= raster::x_res );
         }
#undef IDX

    }// if right
    else  // left
    {

#define IDX 0

         for(; (y0<y1) ;  ++y0,x+= dx)
         {
          span_extents[y0][IDX] = FX1.d(x);
          assert( span_extents[y0][IDX]  >= 0 );
          assert( span_extents[y0][IDX]  <= raster::x_res );
         }
#undef IDX

    }


}


namespace raster
{
 trace_edge_func_t *trace_edge_func = trace_edge_clipped;
}

void __fastcall raster::set_edge_clipping(int do_clipping)
{
 trace_edge_func= do_clipping ? trace_edge_clipped :  trace_edge_unclipped;
}















/*

//
// old version
//

void trace_edge_clipped(float fx0, float fy0, float fx1,float fy1) 
  {  

    int y0 = d_to_int(fy0+0.5f); // round up
    int y1 = d_to_int(fy1+0.5f);
  
    // does line seg exist?
    int dy = y1 - y0;
    if(dy == 0 )
      return;

    float rdx = (fx1 - fx0)/(fy1 - fy0);
	
    // are we a right or a left edge?
    int right;

    if( dy < 0 )
     {
     // we are a right edge
      right = 1;
      dy = -dy;
      swap_obs(fx0,fx1);
      swap_obs(fy0,fy1);
      swap_obs(y0,y1);
     }
    else
     right = 0;

    // trivial y clipping
    if( y0 >= raster::y_res )
    {
       return;
    }
    if( y1 <= 0 )
    {
     return;
    }
    
    if( y0 < 0 ) 
       y0 = 0;
    if( y1 > raster::y_res ) 
       y1 = raster::y_res;

    assert( y1 > y0 );


    // test poly bounds
    if(y0 < y_min) y_min = y0;
    if(y1 > y_max) y_max = y1;


    int dx = d_to_fx16( rdx );

    float pre_step = (float)y0 - fy0;

    int  x    = d_to_fx16( fx0  +  pre_step*rdx ) + (FX1.i()-1);
  



    if( right )
    {
#define X_INSIDE ( x < clip_max_x_fx16  )
#define BOUND_VAL raster::x_res
#define IDX 1

     int inside_0 = X_INSIDE;
     int inside_1 =  fx1  <  clip_max_x_flt;


     //
     // most common case, both inside
     //
      if( inside_0 & inside_1 )
       {
         for(; (y0<y1) ;  ++y0,x+= dx)
         {
          span_extents[y0][IDX] = FX1.d(x);
         }
       }

      // both  oustide
      else if( (inside_0|inside_1) == 0  )
       {
         for(;y0<y1;++y0)
         {
          span_extents[y0][IDX] = BOUND_VAL;
         }
       }
      else if( !inside_0 )
       {
         assert(dx <= 0 );
         for(; ((y0<y1)&&! X_INSIDE ) ;  ++y0,x+= dx)
         {
          span_extents[y0][IDX] = BOUND_VAL;
         }
         for(; (y0<y1) ;  ++y0,x+= dx)
         {
          span_extents[y0][IDX] = FX1.d(x);
         }
       }
      else
      {
         assert(dx >= 0 );
         for(; ((y0<y1)&& X_INSIDE ) ;  ++y0,x+= dx)
         {
          span_extents[y0][IDX] = FX1.d(x);
         }

         for(; (y0<y1) ;  ++y0)
         {
          span_extents[y0][IDX] = BOUND_VAL;
         }

      }
#undef X_INSIDE
#undef BOUND_VAL
#undef IDX

    }// if right
    else  // left
    {

#define X_INSIDE ( x > (FX1.i()/2)  )
#define BOUND_VAL 0
#define IDX 0

     int inside_0 =  X_INSIDE;
     int inside_1 =  fx1  > -0.5f;


     //
     // most common case, both inside
     //
      if( inside_0 & inside_1 )
       {
         for(; (y0<y1) ;  ++y0,x+= dx)
         {
          span_extents[y0][IDX] = FX1.d(x);
         }
       }
      // both  oustide
      else if( (inside_0|inside_1) == 0  )
       {

         for(;y0<y1;++y0)
         {
          span_extents[y0][IDX] = BOUND_VAL;
         }
       }
      else if( !inside_0 )
       {
         assert(dx >= 0 );
         for(; ((y0<y1)&&! X_INSIDE ) ;  ++y0,x+= dx)
         {
          span_extents[y0][IDX] = BOUND_VAL;
         }
         for(; (y0<y1) ;  ++y0,x+= dx)
         {
          span_extents[y0][IDX] = FX1.d(x);
         }
      }
      else
      {
         assert(dx <= 0 );
         for(; ((y0<y1)&& X_INSIDE ) ;  ++y0,x+= dx)
         {
          span_extents[y0][IDX] = FX1.d(x);
         }

         for(; (y0<y1) ;  ++y0)
         {
          span_extents[y0][IDX] = BOUND_VAL;
         }

      }
#undef X_INSIDE
#undef BOUND_VAL
#undef IDX

    }


}
  

*/